/*
 * Copyright (C) 2007-2009 Nokia Corporation.
 * Copyright (c) 2011-2014, NVIDIA CORPORATION.  All rights reserved.
 *
 * Author: Felipe Contreras <felipe.contreras@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "gstomx_base_videoenc.h"
#include "gstomx.h"
#include "gstomx_nvutils.h"
#include <string.h>             /* for strcmp */

#ifdef TEGRA
#include "NVOMX_IndexExtensions.h"
#endif

enum
{
  ARG_0,
  ARG_BITRATE,
  ARG_QUALITY,
  ARG_RC_MODE,
  ARG_INTRA_FRAME_INTERVAL,
  ARG_QUANT_RANGE,
  ARG_TEMPORAL_TRADEOFF,
  ARG_BIT_PACKETIZATION,
  ARG_INSERT_SPSPPSATIDR,
  ARG_FRAMERATE
};

enum
{
  /* actions */
  SIGNAL_FORCE_IDR,
  LAST_SIGNAL
};

#define DEFAULT_BITRATE                  4000000
#define DEFAULT_QUALITY_LEVEL            2
#define DEFAULT_INTRA_FRAME_INTERVAL     60
#define ENCODER_CONF_LOCATION            "/etc/enctune.conf"
#define DEFAULT_RC_MODE                  NVX_VIDEO_RateControlMode_VBR2
#define DEFAULT_TEMPORAL_TRADEOFF_TYPE   GST_30_FPS
#define DEFAULT_BIT_PACKETIZATION        FALSE
#define DEFAULT_INSERT_SPSPPSATIDR       TRUE
#define DEFAULT_FRAMERATE                30

#define GST_TYPE_OMX_VID_ENC_RCMODE (gst_omx_videnc_rc_mode_get_type ())
#define GST_TYPE_OMX_VID_ENC_TEMPORAL_TRADEOFF (gst_omx_videnc_temporal_tradeoff_get_type ())

static guint gst_omx_videoenc_signals[LAST_SIGNAL] = { 0 };

static void gstomx_video_encoder_forceIDR (GstOmxBaseVideoEnc * self);

static GType
gst_omx_videnc_temporal_tradeoff_get_type (void)
{
  static volatile gsize temporal_tradeoff_type_type = 0;
  static const GEnumValue temporal_tradeoff_type[] = {
    {GST_30_FPS, "GST_OMX_VIDENC_FORCE_FRAMERATE_AS_FAST_AS_POSSIBLE", "30fps"},
    {GST_24_FPS, "GST_OMX_VIDENC_FORCE_FRAMERATE_24", "24fps"},
    {GST_20_FPS, "GST_OMX_VIDENC_FORCE_FRAMERATE_20", "20fps"},
    {GST_15_FPS, "GST_OMX_VIDENC_FORCE_FRAMERATE_15", "15fps"},
    {GST_10_FPS, "GST_OMX_VIDENC_FORCE_FRAMERATE_10", "10fps"},
    {0, NULL, NULL}
  };

  if (g_once_init_enter (&temporal_tradeoff_type_type)) {
    GType tmp = g_enum_register_static ("GstOmxVideoEncTemporalTradeoffType",
        temporal_tradeoff_type);
    g_once_init_leave (&temporal_tradeoff_type_type, tmp);
  }

  return (GType) temporal_tradeoff_type_type;
}


static GType
gst_omx_videnc_rc_mode_get_type (void)
{
  static volatile gsize rcmode_type_type = 0;
  static const GEnumValue rcmode_type[] = {
    {NVX_VIDEO_RateControlMode_CBR, "GST_OMX_VIDENC_RCMODE_TYPE_CONSTANT",
        "cbr"},
    {NVX_VIDEO_RateControlMode_VBR, "GST_OMX_VIDENC_RCMODE_TYPE_VARIABLE",
        "vbr"},
    {NVX_VIDEO_RateControlMode_VBR2, "GST_OMX_VIDENC_RCMODE_TYPE_VARIABLE2",
        "vbr2"},
    {0, NULL, NULL}
  };

  if (g_once_init_enter (&rcmode_type_type)) {
    GType tmp =
        g_enum_register_static ("GstOmxVideoEncRCModeType", rcmode_type);
    g_once_init_leave (&rcmode_type_type, tmp);
  }

  return (GType) rcmode_type_type;
}

static gboolean
gst_omx_set_quantization_range (OMX_HANDLETYPE omx_handle, const gchar * arr)
{
  OMX_INDEXTYPE index;
  NVX_CONFIG_VIDENC_QUANTIZATION_RANGE pQuantRange;
  OMX_ERRORTYPE err;
  gint pmin, pmax, imin, imax;
  gchar *str;

  pmin = atoi (arr);
  str = g_strstr_len (arr, -1, "-");
  pmax = atoi (str + 1);
  str = g_strrstr (arr, ":");
  imin = atoi (str + 1);
  str = g_strrstr (arr, "-");
  imax = atoi (str + 1);

  OMX_GetExtensionIndex (omx_handle,
      NVX_INDEX_CONFIG_VIDEO_ENCODE_QUANTIZATION_RANGE, &index);

  G_OMX_INIT_PARAM (pQuantRange);

  pQuantRange.nMinQpP = pmin;
  pQuantRange.nMaxQpP = pmax;
  pQuantRange.nMinQpI = imin;
  pQuantRange.nMaxQpI = imax;

  err = OMX_SetConfig (omx_handle, index, &pQuantRange);
  if (err != OMX_ErrorNone) {
    GST_WARNING ("Can not set Quantization range, Error: %x", err);
    return FALSE;
  }

  return TRUE;
}

static gboolean
gst_omx_get_quantization_range (OMX_HANDLETYPE omx_handle, GValue * value)
{
  OMX_INDEXTYPE index;
  NVX_CONFIG_VIDENC_QUANTIZATION_RANGE pQuantRange;
  OMX_ERRORTYPE err;
  gint pmin, pmax, imin, imax;
  gchar arr[100];

  OMX_GetExtensionIndex (omx_handle,
      NVX_INDEX_CONFIG_VIDEO_ENCODE_QUANTIZATION_RANGE, &index);

  G_OMX_INIT_PARAM (pQuantRange);
  err = OMX_GetConfig (omx_handle, index, &pQuantRange);

  if (err != OMX_ErrorNone) {
    GST_WARNING ("Can not get Quantization range, Error: %x", err);
    return FALSE;
  }

  pmin = pQuantRange.nMinQpP;
  pmax = pQuantRange.nMaxQpP;
  imin = pQuantRange.nMinQpI;
  imax = pQuantRange.nMaxQpI;
  sprintf (arr, "%d-%d:%d-%d", pmin, pmax, imin, imax);

  g_value_set_string (value, arr);
  return TRUE;
}

GSTOMX_BOILERPLATE (GstOmxBaseVideoEnc, gst_omx_base_videoenc, GstOmxBaseFilter,
    GST_OMX_BASE_FILTER_TYPE);

static GstCaps *
generate_sink_template (void)
{
  GstCaps *caps;
  GstStructure *struc, *struc1, *struc2;

  caps = gst_caps_new_empty ();

  struc = gst_structure_new ("video/x-raw-yuv",
      "width", GST_TYPE_INT_RANGE, 16, 4096,
      "height", GST_TYPE_INT_RANGE, 16, 4096,
      "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);

  struc1 = gst_structure_new ("video/x-nv-yuv",
      "width", GST_TYPE_INT_RANGE, 16, 4096,
      "height", GST_TYPE_INT_RANGE, 16, 4096,
      "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);
  struc2 = gst_structure_new ("video/x-nvrm-yuv",
      "width", GST_TYPE_INT_RANGE, 16, 4096,
      "height", GST_TYPE_INT_RANGE, 16, 4096,
      "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);

  {
    GValue list;
    GValue val;

    list.g_type = val.g_type = 0;

    g_value_init (&list, GST_TYPE_LIST);
    g_value_init (&val, GST_TYPE_FOURCC);

    gst_value_set_fourcc (&val, GST_MAKE_FOURCC ('I', '4', '2', '0'));
    gst_value_list_append_value (&list, &val);

#ifndef TEGRA
    gst_value_set_fourcc (&val, GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'));
    gst_value_list_append_value (&list, &val);

    gst_value_set_fourcc (&val, GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'));
    gst_value_list_append_value (&list, &val);
#endif

    gst_value_set_fourcc (&val, GST_MAKE_FOURCC ('N', 'V', '1', '2'));
    gst_value_list_append_value (&list, &val);

    gst_structure_set_value (struc, "format", &list);
    gst_structure_set_value (struc1, "format", &list);
    gst_structure_set_value (struc2, "format", &list);

    g_value_unset (&val);
    g_value_unset (&list);
  }

  gst_caps_append_structure (caps, struc);
  gst_caps_append_structure (caps, struc1);
  gst_caps_append_structure (caps, struc2);

  return caps;
}

static void
type_base_init (gpointer g_class)
{
  GstElementClass *element_class;

  element_class = GST_ELEMENT_CLASS (g_class);

  {
    GstPadTemplate *template;

    template = gst_pad_template_new ("sink", GST_PAD_SINK,
        GST_PAD_ALWAYS, generate_sink_template ());

    gst_element_class_add_pad_template (element_class, template);
  }
}

static OMX_ERRORTYPE
gst_omx_encoder_set_bitrate (GObject * obj, guint bitrate)
{
  GstOmxBaseVideoEnc *self;
  GOmxCore *gomx;
  OMX_VIDEO_CONFIG_BITRATETYPE oBitrate;
  OMX_ERRORTYPE err = OMX_ErrorNone;

  self = GST_OMX_BASE_VIDEOENC (obj);
  gomx = (GOmxCore *) self->omx_base.gomx;

  G_OMX_INIT_PARAM (oBitrate);
  self->bitrate = bitrate;
  OMX_GetConfig (gomx->omx_handle, OMX_IndexConfigVideoBitrate, &oBitrate);
  oBitrate.nEncodeBitrate = self->bitrate;
  if ((err =
          OMX_SetConfig (gomx->omx_handle, OMX_IndexConfigVideoBitrate,
              &oBitrate)) != OMX_ErrorNone) {
    GST_WARNING_OBJECT (gomx->object,
        "Error while setting bitrate config : %x\n", err);
  }

  return err;
}

static void
set_property (GObject * obj,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstOmxBaseVideoEnc *self;
  GstOmxBaseFilter *omx_base;
  OMX_ERRORTYPE err = OMX_ErrorNone;

  self = GST_OMX_BASE_VIDEOENC (obj);
  omx_base = GST_OMX_BASE_FILTER (self);

  switch (prop_id) {
    case ARG_BITRATE:
      gst_omx_encoder_set_bitrate (obj, g_value_get_uint (value));
      break;
    case ARG_QUALITY:
      self->quality_level = g_value_get_uint (value);
      break;
    case ARG_RC_MODE:
      self->rc_mode = g_value_get_enum (value);
      break;
    case ARG_INTRA_FRAME_INTERVAL:
      self->iframeinterval = g_value_get_uint (value);
      break;
    case ARG_QUANT_RANGE:
      gst_omx_set_quantization_range (omx_base->gomx->omx_handle,
          g_value_get_string (value));
      break;
    case ARG_TEMPORAL_TRADEOFF:
      err =
          gstomx_set_video_encoder_temporaltradeoff (omx_base->gomx->omx_handle,
          g_value_get_enum (value));
      if (err == OMX_ErrorNone)
        self->temporal_tradeoff = g_value_get_enum (value);
      else
        g_print ("failed to set temporal tradeoff\n");
      break;
    case ARG_BIT_PACKETIZATION:
      self->bit_packetization = g_value_get_boolean (value);
      break;
    case ARG_INSERT_SPSPPSATIDR:
      self->insert_SPSPPSatIDR = g_value_get_boolean (value);
      break;
    case ARG_FRAMERATE:
      self->framerate_num = g_value_get_int (value);
      self->framerate_denom = 1;
      gstomx_set_video_encoder_framerate (omx_base->gomx->omx_handle,
          self->framerate_num);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
  }
}

static void
get_property (GObject * obj, guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstOmxBaseVideoEnc *self;
  GstOmxBaseFilter *omx_base;

  self = GST_OMX_BASE_VIDEOENC (obj);
  omx_base = GST_OMX_BASE_FILTER (self);

  switch (prop_id) {
    case ARG_BITRATE:
      g_value_set_uint (value, self->bitrate);
      break;
    case ARG_QUALITY:
      g_value_set_uint (value, self->quality_level);
      break;
    case ARG_RC_MODE:
      g_value_set_enum (value, self->rc_mode);
      break;
    case ARG_INTRA_FRAME_INTERVAL:
      g_value_set_uint (value, self->iframeinterval);
      break;
    case ARG_QUANT_RANGE:
      gst_omx_get_quantization_range (omx_base->gomx->omx_handle, value);
      break;
    case ARG_TEMPORAL_TRADEOFF:
      g_value_set_enum (value, self->temporal_tradeoff);
      break;
    case ARG_BIT_PACKETIZATION:
      g_value_set_boolean (value, self->bit_packetization);
      break;
    case ARG_INSERT_SPSPPSATIDR:
      g_value_set_boolean (value, self->insert_SPSPPSatIDR);
      break;
    case ARG_FRAMERATE:
      self->framerate_num =
          gstomx_get_video_encoder_framerate (omx_base->gomx->omx_handle);
      self->framerate_denom = 1;
      g_value_set_int (value, self->framerate_num);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
  }
}

static void
type_class_init (gpointer g_class, gpointer class_data)
{
  GObjectClass *gobject_class;
  GstOmxBaseVideoEnc *self;
  GstOmxBaseVideoEncClass *gstvideoenc_class;

  gobject_class = G_OBJECT_CLASS (g_class);

  self = GST_OMX_BASE_VIDEOENC (g_class);
  gstvideoenc_class = (GstOmxBaseVideoEncClass *) G_OBJECT_CLASS (self);

  /* Properties stuff */
  {
    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    g_object_class_install_property (gobject_class, ARG_BITRATE,
        g_param_spec_uint ("bitrate", "Bit-rate",
            "Encoding bit-rate",
            0, G_MAXUINT, DEFAULT_BITRATE,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_QUALITY,
        g_param_spec_uint ("quality-level", "quality-level",
            "Encoding quality-level",
            0, 2, DEFAULT_QUALITY_LEVEL,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_RC_MODE,
        g_param_spec_enum ("rc-mode", "rc-mode",
            "Encoding rate control mode",
            GST_TYPE_OMX_VID_ENC_RCMODE, DEFAULT_RC_MODE,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_INTRA_FRAME_INTERVAL,
        g_param_spec_uint ("iframeinterval", "Intra Frame interval",
            "Encoding Intra Frame occurance frequency",
            0, G_MAXUINT, DEFAULT_INTRA_FRAME_INTERVAL,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_QUANT_RANGE,
        g_param_spec_string ("qp-range", "qpp-range",
            "Qunatization range for P and I frame", "10-51:5-30",
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_TEMPORAL_TRADEOFF,
        g_param_spec_enum ("temporal-tradeoff", "Temporal Tradeoff for encoder",
            "Temporal Tradeoff value for encoder",
            GST_TYPE_OMX_VID_ENC_TEMPORAL_TRADEOFF,
            DEFAULT_TEMPORAL_TRADEOFF_TYPE,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_BIT_PACKETIZATION,
        g_param_spec_boolean ("bit-packetization", "Bit Based Packetization",
            "Whether or not Packet size is based upon Number Of bits",
            FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_INSERT_SPSPPSATIDR,
        g_param_spec_boolean ("insert-spsppsatidr", "Insert SPS & PPS at IDR",
            "Whether or not insert SPS & PPS at every IDR",
            TRUE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    g_object_class_install_property (gobject_class, ARG_FRAMERATE,
        g_param_spec_int ("framerate", "Framerate",
            "Frames per second",
            0, 60, DEFAULT_FRAMERATE,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

    gst_omx_videoenc_signals[SIGNAL_FORCE_IDR] =
        g_signal_new ("force-IDR",
        G_TYPE_FROM_CLASS (gstvideoenc_class),
        G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
        G_STRUCT_OFFSET (GstOmxBaseVideoEncClass, force_IDR),
        NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

    gstvideoenc_class->force_IDR = gstomx_video_encoder_forceIDR;

  }
}

static gboolean
sink_setcaps (GstPad * pad, GstCaps * caps)
{
  GstStructure *structure;
  GstOmxBaseVideoEnc *self;
  GstOmxBaseFilter *omx_base;
  GOmxCore *gomx;
  OMX_COLOR_FORMATTYPE color_format = OMX_COLOR_FormatUnused;
  gint width = 0;
  gint height = 0;
  const GValue *framerate = NULL;

  self = GST_OMX_BASE_VIDEOENC (GST_PAD_PARENT (pad));
  omx_base = GST_OMX_BASE_FILTER (self);
  gomx = (GOmxCore *) omx_base->gomx;

  omx_base->in_port->buffer_type = BUFFER_TYPE_EGLIMAGE;

  GST_INFO_OBJECT (self, "setcaps (sink): %" GST_PTR_FORMAT, caps);

  g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

  structure = gst_caps_get_structure (caps, 0);

  gst_structure_get_int (structure, "width", &width);
  gst_structure_get_int (structure, "height", &height);

  if (strcmp (gst_structure_get_name (structure), "video/x-raw-yuv") == 0 ||
      strcmp (gst_structure_get_name (structure), "video/x-nv-yuv") == 0 ||
      strcmp (gst_structure_get_name (structure), "video/x-nvrm-yuv") == 0) {
    guint32 fourcc;

    framerate = gst_structure_get_value (structure, "framerate");
    if (framerate) {
      self->framerate_num = gst_value_get_fraction_numerator (framerate);
      self->framerate_denom = gst_value_get_fraction_denominator (framerate);
    }

    if (gst_structure_get_fourcc (structure, "format", &fourcc)) {
      switch (fourcc) {
        case GST_MAKE_FOURCC ('I', '4', '2', '0'):
          color_format = OMX_COLOR_FormatYUV420PackedPlanar;
          break;
        case GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'):
          color_format = OMX_COLOR_FormatYCbYCr;
          break;
        case GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'):
          color_format = OMX_COLOR_FormatCbYCrY;
          break;
        case GST_MAKE_FOURCC ('N', 'V', '1', '2'):
          color_format = OMX_COLOR_FormatYUV420SemiPlanar;
          break;
      }
    }

    if (strcmp (gst_structure_get_name (structure), "video/x-nv-yuv") == 0) {
#ifdef TEGRA
      gstomx_use_nvbuffer_extension (gomx->omx_handle,
          omx_base->in_port->port_index);
#endif
      omx_base->in_port->buffer_type = BUFFER_TYPE_NVBUFFER;
    } else if (strcmp (gst_structure_get_name (structure),
            "video/x-nvrm-yuv") == 0) {
#ifdef TEGRA
      gstomx_use_nvrmsurf_extension (gomx->omx_handle);
#endif
      omx_base->in_port->buffer_type = BUFFER_TYPE_NVRMBUFFER;
    } else if (strcmp (gst_structure_get_name (structure),
            "video/x-raw-yuv") == 0) {
      omx_base->in_port->buffer_type = BUFFER_TYPE_RAW;
    }
    omx_base->out_port->buffer_type = BUFFER_TYPE_RAW;
    omx_base->share_input_buffer = FALSE;
    gomx->share_input_buffer = FALSE;
    omx_base->share_output_buffer = TRUE;
    gomx->share_output_buffer = TRUE;

  }

  OMX_PARAM_PORTDEFINITIONTYPE param;
  G_OMX_INIT_PARAM (param);

  /* Input port configuration. */
  param.nPortIndex = omx_base->in_port->port_index;
  OMX_GetParameter (gomx->omx_handle, OMX_IndexParamPortDefinition, &param);

  param.format.video.nFrameWidth = width;
  param.format.video.nFrameHeight = height;
  param.format.video.eColorFormat = color_format;
  param.format.video.eCompressionFormat = OMX_VIDEO_CodingUnused;

  if (framerate) {
    /* convert to Q.16 */
    param.format.video.xFramerate =
        (gst_value_get_fraction_numerator (framerate) << 16) /
        gst_value_get_fraction_denominator (framerate);
  }

  OMX_SetParameter (gomx->omx_handle, OMX_IndexParamPortDefinition, &param);

  if ((1920 == width) && (height >= 1072 && height <= 1088)) {
    self->quality_level = 1;
  }

  return gst_pad_set_caps (pad, caps);
}

static void
gstomx_video_encoder_forceIDR (GstOmxBaseVideoEnc * self)
{
  GOmxCore *gomx;
  GstOmxBaseFilter *omx_base;

  OMX_ERRORTYPE eError = OMX_ErrorNone;
  OMX_CONFIG_INTRAREFRESHVOPTYPE forceIDR;

  omx_base = GST_OMX_BASE_FILTER (self);
  gomx = (GOmxCore *) omx_base->gomx;

  G_OMX_INIT_PARAM (forceIDR);

  forceIDR.nPortIndex = 1;
  forceIDR.IntraRefreshVOP = OMX_TRUE;

  eError =
      OMX_SetConfig (gomx->omx_handle, OMX_IndexConfigVideoIntraVOPRefresh,
      &forceIDR);
  if (OMX_ErrorNone != eError)
    g_debug ("Failed to force IDR frame\n");
}

static void
omx_setup (GstOmxBaseFilter * omx_base)
{
  GstOmxBaseVideoEnc *self;
  GOmxCore *gomx;
  OMX_ERRORTYPE status = OMX_ErrorNone;

  OMX_IMAGE_PARAM_QFACTORTYPE oQFactor;
  OMX_PARAM_PORTDEFINITIONTYPE param;

  self = GST_OMX_BASE_VIDEOENC (omx_base);
  gomx = (GOmxCore *) omx_base->gomx;

  GST_INFO_OBJECT (omx_base, "begin");

  G_OMX_INIT_PARAM (param);

  //Sets OMX Encoder's Conf File Location
  gchar *location = g_strdup (ENCODER_CONF_LOCATION);
  status = gstomx_set_EncConfLocation (gomx->omx_handle, location);
  if (status != OMX_ErrorNone)
    g_print ("\n%s error while setting EncConfLocation\n", __FUNCTION__);
  g_free (location);

  /* Output port configuration. */
  param.nPortIndex = omx_base->out_port->port_index;
  OMX_GetParameter (gomx->omx_handle, OMX_IndexParamPortDefinition, &param);
  param.format.video.eCompressionFormat = self->compression_format;
  OMX_SetParameter (gomx->omx_handle, OMX_IndexParamPortDefinition, &param);

  gstomx_set_video_encoder_ratecontrolmode (gomx->omx_handle, self->rc_mode);

  gst_omx_encoder_set_bitrate ((GObject *) self, self->bitrate);

  if (self->framerate_denom != 0) {
    gstomx_set_video_encoder_framerate (gomx->omx_handle,
        self->framerate_num / self->framerate_denom);
  }

  G_OMX_INIT_PARAM (oQFactor);

  switch (self->quality_level) {
    case 0:
      oQFactor.nQFactor = 33;
      break;
    case 1:
      oQFactor.nQFactor = 66;
      break;
    case 2:
      oQFactor.nQFactor = 100;
      break;
    default:
      oQFactor.nQFactor = 66;
      break;
  }
  OMX_SetParameter (gomx->omx_handle, OMX_IndexParamQFactor, &oQFactor);

  gstomx_set_video_encoder_temporaltradeoff (gomx->omx_handle,
      self->temporal_tradeoff);

  if (self->ifi_setup != NULL)
    self->ifi_setup (omx_base);

  gstomx_set_video_encoder_property (gomx->omx_handle,
      self->bit_packetization, self->insert_SPSPPSatIDR);

  if (self->configure_output_format) {
    self->configure_output_format (omx_base);
  }

  GST_INFO_OBJECT (omx_base, "end");
}

static void
setup_ports (GstOmxBaseFilter * self)
{
  GOmxCore *core;
  OMX_PARAM_PORTDEFINITIONTYPE param;
  core = self->gomx;

  G_OMX_INIT_PARAM (param);

  /* Input port configuration. */
  param.nPortIndex = 0;
  OMX_GetParameter (core->omx_handle, OMX_IndexParamPortDefinition, &param);
  self->in_port = g_omx_core_setup_port (core, &param);
  gst_pad_set_element_private (self->sinkpad, self->in_port);

  /* Output port configuration. */
  param.nPortIndex = 1;
  OMX_GetParameter (core->omx_handle, OMX_IndexParamPortDefinition, &param);
  self->out_port = g_omx_core_setup_port (core, &param);
  gst_pad_set_element_private (self->srcpad, self->out_port);

  if (self->in_port->buffer_type == BUFFER_TYPE_RAW) {
    if (gstomx_set_video_plane_stride_alignment (core->omx_handle, 4)
        != OMX_ErrorNone) {
      g_warning ("GSTOMX:VIDEOENC: failed to set plane stride align\n");
    }
  }
}

static void
type_instance_init (GTypeInstance * instance, gpointer g_class)
{
  GstOmxBaseFilter *omx_base;
  GstOmxBaseVideoEnc *self;

  omx_base = GST_OMX_BASE_FILTER (instance);
  self = GST_OMX_BASE_VIDEOENC (instance);

  omx_base->omx_setup = omx_setup;
  omx_base->setup_ports = setup_ports;

  gst_pad_set_setcaps_function (omx_base->sinkpad, sink_setcaps);

  self->bitrate = DEFAULT_BITRATE;
  self->rc_mode = DEFAULT_RC_MODE;
  self->quality_level = DEFAULT_QUALITY_LEVEL;
  self->iframeinterval = DEFAULT_INTRA_FRAME_INTERVAL;
  self->temporal_tradeoff = DEFAULT_TEMPORAL_TRADEOFF_TYPE;
  self->bit_packetization = DEFAULT_BIT_PACKETIZATION;
  self->insert_SPSPPSatIDR = DEFAULT_INSERT_SPSPPSATIDR;
  self->framerate_num = DEFAULT_FRAMERATE;
  self->framerate_denom = 1;
}
