/*
 * Copyright (c) 2014, NVIDIA CORPORATION.  All rights reserved.
 *
 * Based on code copyright/by:
 *
 * Author: Felipe Contreras <felipe.contreras@nokia.com>
 * Copyright (C) 2007-2009 Nokia Corporation.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "gstomx_filereadersrc.h"
#include "gstomx_base_src.h"
#include "gstomx.h"

enum
{
  ARG_0,
  ARG_FILE_NAME,
};

typedef struct NVX_PARAM_FILENAME
{
  OMX_U32 nSize;                /**< size of the structure in bytes */
  OMX_VERSIONTYPE nVersion;     /**< NVX extensions specification version information (1.0.0.0)*/
  OMX_STRING pFilename;         /**< name of file as supported by stdio implementation */
} NVX_PARAM_FILENAME;

GSTOMX_BOILERPLATE (GstOmxFilereaderSrc, gst_omx_filereadersrc, GstOmxBaseSrc,
    GST_OMX_BASE_SRC_TYPE);

static GstCaps *
generate_src_template (void)
{
  GstCaps *caps;

  caps = gst_caps_new_any ();

  return caps;
}

static void
type_base_init (gpointer g_class)
{
  GstElementClass *element_class;

  element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details_simple (element_class,
      "OpenMAX IL filereader src element",
      "None", "Does nothing", "Felipe Contreras");

  {
    GstPadTemplate *template;

    template = gst_pad_template_new ("src", GST_PAD_SRC,
        GST_PAD_ALWAYS, generate_src_template ());

    gst_element_class_add_pad_template (element_class, template);
  }
}

static gboolean
setcaps (GstBaseSrc * gst_src, GstCaps * caps)
{
  GstOmxBaseSrc *self;

  self = GST_OMX_BASE_SRC (gst_src);

  GST_INFO_OBJECT (self, "setcaps (src): %" GST_PTR_FORMAT, caps);

  g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

  return TRUE;
}

static void
settings_changed_cb (GOmxCore * core)
{
  GstOmxBaseSrc *omx_base;

  omx_base = core->object;

  GST_DEBUG_OBJECT (omx_base, "settings changed");

    /** @todo properly set the capabilities */
}

static void
setup_ports (GstOmxBaseSrc * base_src)
{
  GOmxCore *gomx;
  GstOmxFilereaderSrc *self;

  self = GST_OMX_FILEREADERSRC (base_src);
  gomx = base_src->gomx;

  /* NVIDIA Specific stuff */
  {
    OMX_INDEXTYPE eIndexParamFilename;
    NVX_PARAM_FILENAME oFilenameParam;
    OMX_GetExtensionIndex (gomx->omx_handle,
        "OMX.Nvidia.index.param.filename", &eIndexParamFilename);
    G_OMX_INIT_PARAM (oFilenameParam);
    oFilenameParam.pFilename = self->file_name;
    OMX_SetParameter (gomx->omx_handle, eIndexParamFilename, &oFilenameParam);
  }
}

static void
set_property (GObject * obj,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstOmxFilereaderSrc *self;

  self = GST_OMX_FILEREADERSRC (obj);

  switch (prop_id) {
    case ARG_FILE_NAME:
      if (self->file_name) {
        g_free (self->file_name);
      }
      self->file_name = g_value_dup_string (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
  }
}

static void
get_property (GObject * obj, guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstOmxFilereaderSrc *self;

  self = GST_OMX_FILEREADERSRC (obj);

  switch (prop_id) {
    case ARG_FILE_NAME:
      g_value_set_string (value, self->file_name);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
  }
}

static void
type_class_init (gpointer g_class, gpointer class_data)
{
  GstBaseSrcClass *gst_base_src_class;
  GObjectClass *gobject_class;

  gst_base_src_class = GST_BASE_SRC_CLASS (g_class);
  gobject_class = G_OBJECT_CLASS (g_class);

  gst_base_src_class->set_caps = setcaps;

  /* Properties stuff */
  {
    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    g_object_class_install_property (gobject_class, ARG_FILE_NAME,
        g_param_spec_string ("file-name", "File name",
            "The input filename to use",
            NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  }
}

static void
type_instance_init (GTypeInstance * instance, gpointer g_class)
{
  GstOmxBaseSrc *omx_base;

  omx_base = GST_OMX_BASE_SRC (instance);

  GST_DEBUG_OBJECT (omx_base, "begin");

  omx_base->setup_ports = setup_ports;

  omx_base->gomx->settings_changed_cb = settings_changed_cb;

  GST_DEBUG_OBJECT (omx_base, "end");
}
