/*
 * Copyright (C) 2007-2009 Nokia Corporation.
 * Copyright (c) 2011-2014, NVIDIA CORPORATION.  All rights reserved.
 * Author: Felipe Contreras <felipe.contreras@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "gstomx_mpeg4enc.h"
#include "gstomx.h"

enum
{
  ARG_0,
  ARG_SLICE_HEADER_SPACING,
};

#define DEFAULT_SLICE_HEADER_SPACING   0

GSTOMX_BOILERPLATE (GstOmxMpeg4Enc, gst_omx_mpeg4enc, GstOmxBaseVideoEnc,
    GST_OMX_BASE_VIDEOENC_TYPE);

static GstCaps *
generate_src_template (void)
{
  GstCaps *caps;

  caps = gst_caps_new_simple ("video/mpeg",
      "width", GST_TYPE_INT_RANGE, 16, 4096,
      "height", GST_TYPE_INT_RANGE, 16, 4096,
      "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
      "mpegversion", G_TYPE_INT, 4,
      "systemstream", G_TYPE_BOOLEAN, FALSE, NULL);

  return caps;
}

static void
type_base_init (gpointer g_class)
{
  GstElementClass *element_class;

  element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details_simple (element_class,
      "OpenMAX IL MPEG-4 video encoder",
      "Codec/Encoder/Video",
      "Encodes video in MPEG-4 format with OpenMAX IL", "Felipe Contreras");

  {
    GstPadTemplate *template;

    template = gst_pad_template_new ("src", GST_PAD_SRC,
        GST_PAD_ALWAYS, generate_src_template ());

    gst_element_class_add_pad_template (element_class, template);
  }
}

static void
set_property (GObject * obj,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstOmxMpeg4Enc *self;
  OMX_ERRORTYPE eError = OMX_ErrorNone;

  self = GST_OMX_MPEG4ENC (obj);

  switch (prop_id) {
    case ARG_SLICE_HEADER_SPACING:
      self->slice_header_spacing = g_value_get_ulong (value);
      eError =
          gstomx_set_mpeg4_encoder_property (self->omx_base.omx_base.gomx->
          omx_handle, self->slice_header_spacing,
          self->omx_base.iframeinterval);
      if (eError != OMX_ErrorNone)
        g_print ("Failed setting slice_header_spacing error: %d\n", eError);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
  }
}

static void
get_property (GObject * obj, guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstOmxMpeg4Enc *self;

  self = GST_OMX_MPEG4ENC (obj);

  switch (prop_id) {
    case ARG_SLICE_HEADER_SPACING:
      g_value_set_ulong (value, self->slice_header_spacing);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
  }
}

static void
type_class_init (gpointer g_class, gpointer class_data)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (g_class);

  parent_class = g_type_class_ref (GST_OMX_BASE_VIDEOENC_TYPE);

  {
    gobject_class->set_property = set_property;
    gobject_class->get_property = get_property;

    g_object_class_install_property (gobject_class, ARG_SLICE_HEADER_SPACING,
        g_param_spec_ulong ("slice-header-spacing", "Slice Header Spacing",
            "Slice Header Spacing number of macroblocks/bits in one packet",
            0, G_MAXULONG, DEFAULT_SLICE_HEADER_SPACING,
            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  }
}

static void
settings_changed_cb (GOmxCore * core)
{
  GstOmxBaseVideoEnc *omx_base;
  GstOmxBaseFilter *omx_base_filter;
  guint width;
  guint height;

  omx_base_filter = core->object;
  omx_base = GST_OMX_BASE_VIDEOENC (omx_base_filter);

  GST_DEBUG_OBJECT (omx_base, "settings changed");

  {
    OMX_PARAM_PORTDEFINITIONTYPE param;

    G_OMX_INIT_PARAM (param);

    param.nPortIndex = omx_base_filter->out_port->port_index;
    OMX_GetParameter (core->omx_handle, OMX_IndexParamPortDefinition, &param);

    width = param.format.video.nFrameWidth;
    height = param.format.video.nFrameHeight;
  }

  {
    GstCaps *new_caps;

    new_caps = gst_caps_new_simple ("video/mpeg",
        "mpegversion", G_TYPE_INT, 4,
        "width", G_TYPE_INT, width,
        "height", G_TYPE_INT, height,
        "framerate", GST_TYPE_FRACTION,
        omx_base->framerate_num, omx_base->framerate_denom,
        "systemstream", G_TYPE_BOOLEAN, FALSE, NULL);

    GST_INFO_OBJECT (omx_base, "caps are: %" GST_PTR_FORMAT, new_caps);
    gst_pad_set_caps (omx_base_filter->srcpad, new_caps);
  }
}

static void
ifi_setup (GstOmxBaseFilter * omx_base)
{
  GstOmxBaseVideoEnc *self;
  GOmxCore *gomx;
  OMX_VIDEO_PARAM_MPEG4TYPE oMpeg4Type;

  self = GST_OMX_BASE_VIDEOENC (omx_base);
  gomx = (GOmxCore *) omx_base->gomx;

  G_OMX_INIT_PARAM (oMpeg4Type);
  oMpeg4Type.nPortIndex = omx_base->in_port->port_index;

  OMX_GetParameter (gomx->omx_handle, OMX_IndexParamVideoMpeg4, &oMpeg4Type);
  if (self->iframeinterval != 0) {
    oMpeg4Type.bGov = OMX_TRUE;
    oMpeg4Type.nPFrames = self->iframeinterval - 1;
  }
  OMX_SetParameter (gomx->omx_handle, OMX_IndexParamVideoMpeg4, &oMpeg4Type);
}

static void
type_instance_init (GTypeInstance * instance, gpointer g_class)
{
  GstOmxBaseFilter *omx_base_filter;
  GstOmxBaseVideoEnc *omx_base;
  GstOmxMpeg4Enc *self;

  omx_base_filter = GST_OMX_BASE_FILTER (instance);
  omx_base = GST_OMX_BASE_VIDEOENC (instance);
  self = GST_OMX_MPEG4ENC (instance);

  self->slice_header_spacing = DEFAULT_SLICE_HEADER_SPACING;

  omx_base->compression_format = OMX_VIDEO_CodingMPEG4;
  omx_base->ifi_setup = ifi_setup;

  omx_base_filter->gomx->settings_changed_cb = settings_changed_cb;
}
