/*
 * Copyright (C) 2006-2007 Texas Instruments, Incorporated
 * Copyright (C) 2007-2009 Nokia Corporation.
 * Copyright (c) 2009-2014, NVIDIA CORPORATION.  All rights reserved.
 *
 * Author: Felipe Contreras <felipe.contreras@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "gstomx_util.h"
#include "config.h"
#include <dlfcn.h>

#include "gstomx.h"

GST_DEBUG_CATEGORY (gstomx_util_debug);

/*
 * Forward declarations
 */

static inline void change_state (GOmxCore * core, OMX_STATETYPE state);

static inline void wait_for_state (GOmxCore * core, OMX_STATETYPE state);

static inline void
in_port_cb (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer);

static inline void
out_port_cb (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer);

static inline void
got_buffer (GOmxCore * core,
    GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer);

static OMX_ERRORTYPE
EventHandler (OMX_HANDLETYPE omx_handle,
    OMX_PTR app_data,
    OMX_EVENTTYPE event, OMX_U32 data_1, OMX_U32 data_2, OMX_PTR event_data);

static OMX_ERRORTYPE
EmptyBufferDone (OMX_HANDLETYPE omx_handle,
    OMX_PTR app_data, OMX_BUFFERHEADERTYPE * omx_buffer);

static OMX_ERRORTYPE
FillBufferDone (OMX_HANDLETYPE omx_handle,
    OMX_PTR app_data, OMX_BUFFERHEADERTYPE * omx_buffer);

static inline const char *omx_state_to_str (OMX_STATETYPE omx_state);

static inline const char *omx_error_to_str (OMX_ERRORTYPE omx_error);

GOmxPort *g_omx_core_get_port (GOmxCore * core, guint index);

static void core_deinit (GOmxCore * core);

static inline void port_free_buffers (GOmxPort * port);

static inline void port_allocate_buffers (GOmxPort * port);

static inline void port_start_buffers (GOmxPort * port);

static OMX_CALLBACKTYPE callbacks =
    { EventHandler, EmptyBufferDone, FillBufferDone };

/* protect implementations hash_table */
static GMutex *imp_mutex;
static GHashTable *implementations;
static gboolean initialized;

/*
 * Util
 */

static void
g_ptr_array_clear (GPtrArray * array)
{
  guint index;
  for (index = 0; index < array->len; index++)
    array->pdata[index] = NULL;
}

static void
g_ptr_array_insert (GPtrArray * array, guint index, gpointer data)
{
  if (index + 1 > array->len) {
    g_ptr_array_set_size (array, index + 1);
  }

  array->pdata[index] = data;
}

typedef void (*GOmxPortFunc) (GOmxPort * port);

static inline void
core_for_each_port (GOmxCore * core, GOmxPortFunc func)
{
  guint index;

  for (index = 0; index < core->ports->len; index++) {
    GOmxPort *port;

    port = g_omx_core_get_port (core, index);

    if (port)
      func (port);
  }
}

/*
 * Main
 */

static GOmxImp *imp_new (const gchar * name);
static void imp_free (GOmxImp * imp);

static GOmxImp *
imp_new (const gchar * name)
{
  GOmxImp *imp;

  imp = g_new0 (GOmxImp, 1);

  /* Load the OpenMAX IL symbols */
  {
    void *handle;

    GST_DEBUG ("loading: %s", name);

    imp->dl_handle = handle = dlopen (name, RTLD_LAZY);

    GST_DEBUG ("dlopen(%s) -> %p", name, handle);

    if (!handle) {
      g_warning ("%s\n", dlerror ());
      g_free (imp);
      return NULL;
    }

    imp->mutex = g_mutex_new ();
    imp->sym_table.init = dlsym (handle, "OMX_Init");
    imp->sym_table.deinit = dlsym (handle, "OMX_Deinit");
    imp->sym_table.get_handle = dlsym (handle, "OMX_GetHandle");
    imp->sym_table.free_handle = dlsym (handle, "OMX_FreeHandle");
  }

  return imp;
}

static void
imp_free (GOmxImp * imp)
{
  if (imp->dl_handle) {
    dlclose (imp->dl_handle);
  }
  g_mutex_free (imp->mutex);
  g_free (imp);
}

static inline GOmxImp *
request_imp (const gchar * name)
{
  GOmxImp *imp = NULL;

  g_mutex_lock (imp_mutex);
  imp = g_hash_table_lookup (implementations, name);
  if (!imp) {
    imp = imp_new (name);
    if (imp)
      g_hash_table_insert (implementations, g_strdup (name), imp);
  }
  g_mutex_unlock (imp_mutex);

  if (!imp)
    return NULL;

  g_mutex_lock (imp->mutex);
  if (imp->client_count == 0) {
    OMX_ERRORTYPE omx_error;
    omx_error = imp->sym_table.init ();
    if (omx_error) {
      g_mutex_unlock (imp->mutex);
      return NULL;
    }
  }
  imp->client_count++;
  g_mutex_unlock (imp->mutex);

  return imp;
}

static inline void
release_imp (GOmxImp * imp)
{
  g_mutex_lock (imp->mutex);
  imp->client_count--;
  if (imp->client_count == 0) {
    imp->sym_table.deinit ();
  }
  g_mutex_unlock (imp->mutex);
}

void
g_omx_init (void)
{
  if (!initialized) {
    /* safe as plugin_init is safe */
    imp_mutex = g_mutex_new ();
#ifdef USE_EGLIMAGE
    egl_init ();
#endif
    implementations = g_hash_table_new_full (g_str_hash,
        g_str_equal, g_free, (GDestroyNotify) imp_free);
    initialized = TRUE;
  }
}

void
g_omx_deinit (void)
{
  if (initialized) {
    g_hash_table_destroy (implementations);
    g_mutex_free (imp_mutex);
#ifdef USE_EGLIMAGE
    egl_destroy ();
#endif
    initialized = FALSE;
  }
}

/*
 * Core
 */

GOmxCore *
g_omx_core_new (void *object)
{
  GOmxCore *core;

  core = g_new0 (GOmxCore, 1);

  core->object = object;
  core->ports = g_ptr_array_new ();

  core->omx_state_condition = g_cond_new ();
  core->omx_state_mutex = g_mutex_new ();

  core->done_sem = g_sem_new ();
  core->flush_sem = g_sem_new ();
  core->port_sem = g_sem_new ();

  core->omx_state = OMX_StateInvalid;
  core->share_input_buffer = FALSE;
  core->share_output_buffer = FALSE;

  return core;
}


void
g_omx_core_ports_free (GOmxCore * core)
{
  core_for_each_port (core, g_omx_port_free);
  g_ptr_array_clear (core->ports);
}


void
g_omx_core_free (GOmxCore * core)
{
  core_deinit (core);

  g_sem_free (core->port_sem);
  g_sem_free (core->flush_sem);
  g_sem_free (core->done_sem);

  g_mutex_free (core->omx_state_mutex);
  g_cond_free (core->omx_state_condition);

  g_ptr_array_free (core->ports, TRUE);

  g_free (core);
}

void
g_omx_core_init (GOmxCore * core)
{
  GST_DEBUG_OBJECT (core->object, "loading: %s %s (%s)",
      core->component_name,
      core->component_role ? core->component_role : "", core->library_name);

  core->imp = request_imp (core->library_name);

  if (!core->imp)
    return;

  core->omx_error = core->imp->sym_table.get_handle (&core->omx_handle,
      (char *) core->component_name, core, &callbacks);

  GST_DEBUG_OBJECT (core->object, "OMX_GetHandle(&%p) -> %d",
      core->omx_handle, core->omx_error);

  if (!core->omx_error) {
    core->omx_state = OMX_StateLoaded;

    if (core->component_role) {
      OMX_PARAM_COMPONENTROLETYPE param;

      GST_DEBUG_OBJECT (core->object, "setting component role: %s",
          core->component_role);

      G_OMX_INIT_PARAM (param);

      strncpy ((char *) param.cRole, core->component_role,
          OMX_MAX_STRINGNAME_SIZE);

      OMX_SetParameter (core->omx_handle, OMX_IndexParamStandardComponentRole,
          &param);
    }
  }
}

static void
core_deinit (GOmxCore * core)
{
  if (!core->imp)
    return;

  if (core->omx_state == OMX_StateLoaded || core->omx_state == OMX_StateInvalid) {
    if (core->omx_handle) {
      core->omx_error = core->imp->sym_table.free_handle (core->omx_handle);
      GST_DEBUG_OBJECT (core->object, "OMX_FreeHandle(%p) -> %d",
          core->omx_handle, core->omx_error);
    }
  } else {
    GST_WARNING_OBJECT (core->object, "Incorrect state: %s",
        omx_state_to_str (core->omx_state));
  }

  g_free (core->library_name);
  g_free (core->component_name);
  g_free (core->component_role);

  release_imp (core->imp);
  core->imp = NULL;
}

void
g_omx_core_prepare (GOmxCore * core)
{
  change_state (core, OMX_StateIdle);

  /* Allocate buffers. */
  core_for_each_port (core, port_allocate_buffers);

  wait_for_state (core, OMX_StateIdle);
}

void
g_omx_core_start (GOmxCore * core)
{
  change_state (core, OMX_StateExecuting);
  wait_for_state (core, OMX_StateExecuting);

  if (core->omx_state == OMX_StateExecuting)
    core_for_each_port (core, port_start_buffers);
}

void
g_omx_core_stop (GOmxCore * core)
{
  if (core->omx_state == OMX_StateExecuting ||
      core->omx_state == OMX_StatePause) {
    change_state (core, OMX_StateIdle);
    wait_for_state (core, OMX_StateIdle);
  }
}

void
g_omx_core_pause (GOmxCore * core)
{
  change_state (core, OMX_StatePause);
  wait_for_state (core, OMX_StatePause);
}

void
g_omx_core_unload (GOmxCore * core)
{
  if (core->omx_state == OMX_StateIdle ||
      core->omx_state == OMX_StateWaitForResources ||
      core->omx_state == OMX_StateInvalid) {
    if (core->omx_state != OMX_StateInvalid)
      change_state (core, OMX_StateLoaded);

    core_for_each_port (core, port_free_buffers);

    if (core->omx_state != OMX_StateInvalid)
      wait_for_state (core, OMX_StateLoaded);
  }

  core_for_each_port (core, g_omx_port_free);
  g_ptr_array_clear (core->ports);
}


void
g_omx_core_finish (GOmxCore * core)
{
  if (core->omx_state == OMX_StateIdle ||
      core->omx_state == OMX_StateWaitForResources ||
      core->omx_state == OMX_StateInvalid) {
    if (core->omx_state != OMX_StateInvalid)
      change_state (core, OMX_StateLoaded);

    core_for_each_port (core, port_free_buffers);

    if (core->omx_state != OMX_StateInvalid)
      wait_for_state (core, OMX_StateLoaded);
  }
}


GOmxPort *
g_omx_core_setup_port (GOmxCore * core, OMX_PARAM_PORTDEFINITIONTYPE * omx_port)
{
  GOmxPort *port;
  guint index;

  index = omx_port->nPortIndex;
  port = g_omx_core_get_port (core, index);

  if (!port) {
    port = g_omx_port_new (core, index);
    g_ptr_array_insert (core->ports, index, port);
  }

  g_omx_port_setup (port);

  return port;
}

GOmxPort *
g_omx_core_get_port (GOmxCore * core, guint index)
{
  if (G_LIKELY (index < core->ports->len)) {
    return g_ptr_array_index (core->ports, index);
  }

  return NULL;
}

GOmxPort *
g_omx_core_new_port (GOmxCore * core, guint index)
{
  GOmxPort *port = g_omx_core_get_port (core, index);

  if (port) {
    GST_WARNING_OBJECT (core->object, "port %d already exists", index);
    return port;
  }

  port = g_omx_port_new (core, index);
  g_ptr_array_insert (core->ports, index, port);

  return port;
}

void
g_omx_core_set_done (GOmxCore * core)
{
  g_sem_up (core->done_sem);
}

void
g_omx_core_wait_for_done (GOmxCore * core)
{
  g_sem_down (core->done_sem);
}

void
g_omx_core_flush_start (GOmxCore * core)
{
  core_for_each_port (core, g_omx_port_pause);
}

void
g_omx_core_flush_stop (GOmxCore * core)
{
  core_for_each_port (core, g_omx_port_flush);
  core_for_each_port (core, g_omx_port_resume);
}

/*
 * Port
 */

/**
 * note: this is not intended to be called directly by elements (which should
 * instead use g_omx_core_new_port())
 */
GOmxPort *
g_omx_port_new (GOmxCore * core, guint index)
{
  GOmxPort *port;
  port = g_new0 (GOmxPort, 1);

  port->core = core;
  port->port_index = index;
  port->num_buffers = 0;
  port->buffer_size = 0;
  port->buffers = NULL;

  port->enabled = TRUE;
  port->queue = async_queue_new ();
  port->squeue = async_queue_new ();
  port->mutex = g_mutex_new ();
  port->buffer_type = BUFFER_TYPE_RAW;
#ifdef USE_EGLIMAGE
  port->eglbuffers = NULL;
#endif
#ifdef GST_KPI_MEASURE
  port->first_buffer = FALSE;
#endif

  return port;
}

void
g_omx_port_free (GOmxPort * port)
{
  g_mutex_free (port->mutex);
  async_queue_free (port->queue);
  async_queue_free (port->squeue);
  if (port->buffers)
    g_free (port->buffers);
#ifdef GST_KPI_MEASURE
  if (port->start_timer != NULL) {
    g_timer_destroy (port->start_timer);
    port->start_timer = NULL;
  }
#endif
  g_free (port);
}

void
g_omx_port_setup (GOmxPort * port)
{
  GOmxPortType type = -1;
  OMX_PARAM_PORTDEFINITIONTYPE param;

  G_OMX_INIT_PARAM (param);

  param.nPortIndex = port->port_index;
  OMX_GetParameter (port->core->omx_handle, OMX_IndexParamPortDefinition,
      &param);

  switch (param.eDir) {
    case OMX_DirInput:
      type = GOMX_PORT_INPUT;
      break;
    case OMX_DirOutput:
      type = GOMX_PORT_OUTPUT;
      break;
    default:
      break;
  }

  port->type = type;
    /** @todo should it be nBufferCountMin? */
  port->num_buffers = param.nBufferCountActual;
  port->buffer_size = param.nBufferSize;

  GST_DEBUG_OBJECT (port->core->object,
      "type=%d, num_buffers=%d, buffer_size=%ld, port_index=%d",
      port->type, port->num_buffers, port->buffer_size, port->port_index);

//  g_free (port->buffers);
  async_queue_flush (port->queue);
  port->buffers = g_new0 (OMX_BUFFERHEADERTYPE *, port->num_buffers);
  g_omx_port_resume (port);
  port->enabled = TRUE;
  port->free_buffers = FALSE;
  port->reconfigure_buffers = FALSE;
  port->keep_buffers = FALSE;
  port->segment_sent = FALSE;
  port->buffers_sent = 0;
}


void
g_omx_port_setup_full (GOmxPort * port, guint count, guint size)
{
  OMX_PARAM_PORTDEFINITIONTYPE param;

  G_OMX_INIT_PARAM (param);

  param.nPortIndex = port->port_index;
  OMX_GetParameter (port->core->omx_handle, OMX_IndexParamPortDefinition,
      &param);

  if (count)
    param.nBufferCountActual = count;
  if (size)
    param.nBufferSize = size;
  OMX_SetParameter (port->core->omx_handle, OMX_IndexParamPortDefinition,
      &param);

  return g_omx_port_setup (port);
}


static void
port_allocate_buffers (GOmxPort * port)
{
  guint i;
  gsize size;

#ifdef USE_EGLIMAGE
  /* create eglbuffers on output port, while input port is sharing it. */
  if (port->buffer_type == BUFFER_TYPE_EGLIMAGE) {
    OMX_PARAM_PORTDEFINITIONTYPE param;
    OMX_ERRORTYPE err = OMX_ErrorNone;
    int width, height;
    G_OMX_INIT_PARAM (param);
    param.nPortIndex = 1;
    OMX_GetParameter (port->core->omx_handle, OMX_IndexParamPortDefinition,
        &param);
    width = param.format.video.nFrameWidth;
    height = param.format.video.nFrameHeight;
    if (port->type == GOMX_PORT_OUTPUT) {
      port->eglbuffers = g_new0 (nvomx_eglbuffer *, port->num_buffers);
      for (i = 0; i < port->num_buffers; i++) {
        err =
            nvomx_allocate_eglimages (port->core->display_data,
            &port->eglbuffers[i], width, height);
        if (err != OMX_ErrorNone)
          g_omx_port_disable (port);
        err = OMX_UseEGLImage (port->core->omx_handle, &port->buffers[i],
            port->port_index, NULL, port->eglbuffers[i]->hEglImg);
        if (err != OMX_ErrorNone)
          g_omx_port_disable (port);
      }
    } else {
      for (i = 0; i < port->num_buffers; i++) {
        err = OMX_UseEGLImage (port->core->omx_handle, &port->buffers[i],
            port->port_index, NULL, NULL);
        if (err != OMX_ErrorNone)
          g_omx_port_disable (port);
      }
    }
    return;
  }
#endif //USE_EGLIMAGE
  size = port->buffer_size;

  for (i = 0; i < port->num_buffers; i++) {
    if (port->omx_allocate) {
      GST_DEBUG_OBJECT (port->core->object,
          "%d: OMX_AllocateBuffer(), size=%" G_GSIZE_FORMAT, i, size);

      OMX_AllocateBuffer (port->core->omx_handle, &port->buffers[i],
          port->port_index, NULL, size);
    } else {
      gpointer buffer_data;
      buffer_data = g_malloc (size);
      GST_DEBUG_OBJECT (port->core->object,
          "%d: OMX_UseBuffer(), size=%" G_GSIZE_FORMAT, i, size);
      OMX_UseBuffer (port->core->omx_handle, &port->buffers[i],
          port->port_index, NULL, size, buffer_data);
    }
  }
}

static void
port_free_buffers (GOmxPort * port)
{
  guint i;

  async_queue_flush (port->queue);

#ifdef USE_EGLIMAGE
  if (port->buffer_type == BUFFER_TYPE_EGLIMAGE) {
    for (i = 0; i < port->num_buffers; i++) {
      OMX_FreeBuffer (port->core->omx_handle, port->port_index,
          port->buffers[i]);
      if (port->type == GOMX_PORT_OUTPUT) {
        nvomx_destroy_eglimages (port->core->display_data, port->eglbuffers[i]);
      }
    }
    if (port->type == GOMX_PORT_OUTPUT) {
      g_free (port->eglbuffers);
    }
    return;
  }
#endif // USE_EGLIMAGE

  if (port->buffers) {
    for (i = 0; i < port->num_buffers; i++) {
      OMX_BUFFERHEADERTYPE *omx_buffer;

      omx_buffer = port->buffers[i];

      if (omx_buffer) {
        if (!port->omx_allocate) {
          if (!(port->core->share_input_buffer && port->type == GOMX_PORT_INPUT)
              && !(port->core->share_output_buffer
                  && port->type == GOMX_PORT_OUTPUT)) {
            g_free (omx_buffer->pBuffer);
            omx_buffer->pBuffer = NULL;
          } else {
            if (omx_buffer->pAppPrivate) {
              int refcount;
              GstBuffer *buf = omx_buffer->pAppPrivate;
              refcount = buf->mini_object.refcount;
              while (refcount--)
                gst_buffer_unref (buf);
            } else if (omx_buffer->pBuffer) {
              /* If this is malloced */
              g_free (omx_buffer->pBuffer);
              omx_buffer->pBuffer = NULL;
            }
          }

        }

        OMX_FreeBuffer (port->core->omx_handle, port->port_index, omx_buffer);
        port->buffers[i] = NULL;
      }
    }

    port->num_buffers = 0;
    g_free (port->buffers);
  }

  port->buffers = NULL;
}

static void
port_start_buffers (GOmxPort * port)
{
  guint i;

  for (i = 0; i < port->num_buffers; i++) {
    OMX_BUFFERHEADERTYPE *omx_buffer;

    omx_buffer = port->buffers[i];

    /* If it's an input port we will need to fill the buffer, so put it in
     * the queue, otherwise send to omx for processing (fill it up). */
    if (port->type == GOMX_PORT_INPUT)
      got_buffer (port->core, port, omx_buffer);
    else {
      memset (omx_buffer->pBuffer, 0, omx_buffer->nAllocLen);
      g_omx_port_release_buffer (port, omx_buffer);
      g_atomic_int_inc (&port->buffers_sent);
    }
  }
}

void
g_omx_port_push_buffer (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer)
{
  async_queue_push (port->queue, omx_buffer);
}

void
g_omx_port_store_buffer (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer)
{
  async_queue_push (port->squeue, omx_buffer);
}

OMX_BUFFERHEADERTYPE *
g_omx_port_request_buffer (GOmxPort * port)
{
  return async_queue_pop (port->queue);
}

void
g_omx_port_release_buffer (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer)
{
  if (G_UNLIKELY (port->keep_buffers)) {
    g_omx_port_store_buffer (port, omx_buffer);
    if (port->fhandler)
      port->fhandler (port, port->udata, port->udata2);
  } else {
    switch (port->type) {
      case GOMX_PORT_INPUT:
        OMX_EmptyThisBuffer (port->core->omx_handle, omx_buffer);
        break;
      case GOMX_PORT_OUTPUT:
        if (G_LIKELY ((OMX_FillThisBuffer (port->core->omx_handle,
                        omx_buffer)) == OMX_ErrorNone))
          g_atomic_int_dec_and_test (&port->buffers_sent);
        break;
      default:
        break;
    }
  }
}

static void
port_cleanup (GOmxPort * port)
{
  OMX_BUFFERHEADERTYPE *omx_buffer;
  while ((omx_buffer = async_queue_pop_forced (port->queue))) {
    omx_buffer->nFilledLen = 0;
    g_omx_port_release_buffer (port, omx_buffer);
  }
}

void
g_omx_port_resume (GOmxPort * port)
{
  async_queue_enable (port->queue);
}

void
g_omx_port_pause (GOmxPort * port)
{
  async_queue_disable (port->queue);
}

void
g_omx_port_flush (GOmxPort * port)
{
  if (port->is_audio_port) {
    change_state (port->core, OMX_StatePause);
    wait_for_state (port->core, OMX_StatePause);
  }

  if (port->type == GOMX_PORT_OUTPUT) {
    OMX_SendCommand (port->core->omx_handle, OMX_CommandFlush, port->port_index,
        NULL);
    g_sem_down (port->core->flush_sem);
    port_cleanup (port);
  } else {
    OMX_SendCommand (port->core->omx_handle, OMX_CommandFlush, port->port_index,
        NULL);
    g_sem_down (port->core->flush_sem);
  }

  if (port->is_audio_port) {
    change_state (port->core, OMX_StateExecuting);
    wait_for_state (port->core, OMX_StateExecuting);
  }
}

void
g_omx_port_enable (GOmxPort * port)
{
  GOmxCore *core;

  core = port->core;

  OMX_SendCommand (core->omx_handle, OMX_CommandPortEnable, port->port_index,
      NULL);
  port_allocate_buffers (port);

  g_sem_down (core->port_sem);

  if (core->omx_state != OMX_StateLoaded)
    port_start_buffers (port);
  g_omx_port_resume (port);
}

void
g_omx_port_disable (GOmxPort * port)
{
  GOmxCore *core;

  core = port->core;

  OMX_SendCommand (core->omx_handle, OMX_CommandPortDisable, port->port_index,
      NULL);
  g_omx_port_pause (port);
  port_cleanup (port);
  port_free_buffers (port);

  g_sem_down (core->port_sem);
}

void
g_omx_port_finish (GOmxPort * port)
{
  port->enabled = FALSE;
  async_queue_disable (port->queue);
}

/*
 * Helper functions.
 */

static inline void
change_state (GOmxCore * core, OMX_STATETYPE state)
{
  GST_DEBUG_OBJECT (core->object, "state=%d", state);
  OMX_SendCommand (core->omx_handle, OMX_CommandStateSet, state, NULL);
}

static inline void
complete_change_state (GOmxCore * core, OMX_STATETYPE state)
{
  g_mutex_lock (core->omx_state_mutex);

  core->omx_state = state;
  g_cond_signal (core->omx_state_condition);
  GST_DEBUG_OBJECT (core->object, "state=%d", state);

  g_mutex_unlock (core->omx_state_mutex);
}

static inline void
wait_for_state (GOmxCore * core, OMX_STATETYPE state)
{
  GTimeVal tv;
  gboolean signaled;

  g_mutex_lock (core->omx_state_mutex);

  if (core->omx_error != OMX_ErrorNone) {
    OMX_STATETYPE curstate;
    int numtries = 0;

    /* Unlock the mutex First, Or else it may hang inside EventHandler */
    g_mutex_unlock (core->omx_state_mutex);

    OMX_GetState (core->omx_handle, &curstate);

    /* State may be even InValid, so check that Atleast State change happens */
    while (state != curstate) {
      OMX_GetState (core->omx_handle, &curstate);
      g_usleep (1000);
      /*Just try for 500ms and quit */
      if (curstate == OMX_StateInvalid || numtries > 500) {
        if (curstate == OMX_StateInvalid) {
          GST_ERROR_OBJECT (core->object,
              "openmax returned Invalid State during state transition");
        }

        if (numtries > 500) {
          GST_ERROR_OBJECT (core->object,
              "openmax timed out for state transition");
        }
        break;
      }
      numtries++;
    }
    return;
  }

  g_get_current_time (&tv);
  g_time_val_add (&tv, 15 * G_USEC_PER_SEC);

  /* try once */
  if (core->omx_state != state) {
    signaled =
        g_cond_timed_wait (core->omx_state_condition, core->omx_state_mutex,
        &tv);

    if (!signaled) {
      GST_ERROR_OBJECT (core->object, "timed out switching from '%s' to '%s'",
          omx_state_to_str (core->omx_state), omx_state_to_str (state));
    }
  }

  if (core->omx_error != OMX_ErrorNone)
    goto leave;

  if (core->omx_state != state) {
    GST_ERROR_OBJECT (core->object,
        "wrong state received: state=%d, expected=%d", core->omx_state, state);
  }

leave:
  g_mutex_unlock (core->omx_state_mutex);
}

/*
 * Callbacks
 */

static inline void
in_port_cb (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer)
{
    /** @todo remove this */

  if (!port->enabled)
    return;
}

static inline void
out_port_cb (GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer)
{
    /** @todo remove this */

  if (!port->enabled)
    return;

#if 0
  if (omx_buffer->nFlags & OMX_BUFFERFLAG_EOS) {
    g_omx_port_set_done (port);
    return;
  }
#endif
}

static inline void
got_buffer (GOmxCore * core, GOmxPort * port, OMX_BUFFERHEADERTYPE * omx_buffer)
{
  if (G_UNLIKELY (!omx_buffer)) {
    return;
  }

  if (G_UNLIKELY (port->free_buffers)) {
    if (!port->omx_allocate) {
      if (!(port->core->share_input_buffer && port->type == GOMX_PORT_INPUT) &&
          !(port->core->share_output_buffer
              && port->type == GOMX_PORT_OUTPUT)) {
        g_free (omx_buffer->pBuffer);
        omx_buffer->pBuffer = NULL;

      } else {
        if (omx_buffer->pAppPrivate) {
          int refcount;
          GstBuffer *buf = omx_buffer->pAppPrivate;
          refcount = buf->mini_object.refcount;
          while (refcount--)
            gst_buffer_unref (buf);
        } else if (omx_buffer->pBuffer) {
          /* If this is malloced */
          g_free (omx_buffer->pBuffer);
          omx_buffer->pBuffer = NULL;
        }
      }
    }

    OMX_FreeBuffer (port->core->omx_handle, port->port_index, omx_buffer);
  } else if (G_LIKELY (port)) {
    if (!port->reconfigure_buffers)
      g_omx_port_push_buffer (port, omx_buffer);

    switch (port->type) {
      case GOMX_PORT_INPUT:
        in_port_cb (port, omx_buffer);
        break;
      case GOMX_PORT_OUTPUT:
        out_port_cb (port, omx_buffer);
        break;
      default:
        break;
    }
  }
}


void
g_omx_port_set_active (GOmxPort * port, gboolean val, gboolean free)
{
  if (val) {
    OMX_SendCommand (port->core->omx_handle, OMX_CommandPortEnable,
        port->port_index, NULL);
  } else {
    port->free_buffers = free;
    OMX_SendCommand (port->core->omx_handle, OMX_CommandPortDisable,
        port->port_index, NULL);
  }

  g_sem_down (port->core->port_sem);

  if (!val) {
    OMX_BUFFERHEADERTYPE *omx_buffer;

    while ((omx_buffer = async_queue_pop_forced (port->queue))) {
      if (free) {
        if (!port->omx_allocate) {
          if (!(port->core->share_input_buffer && port->type == GOMX_PORT_INPUT)
              && !(port->core->share_output_buffer
                  && port->type == GOMX_PORT_OUTPUT)) {
            g_free (omx_buffer->pBuffer);
            omx_buffer->pBuffer = NULL;

          } else {
            if (omx_buffer->pAppPrivate) {
              int refcount;
              GstBuffer *buf = omx_buffer->pAppPrivate;
              refcount = buf->mini_object.refcount;
              while (refcount--)
                gst_buffer_unref (buf);
            } else if (omx_buffer->pBuffer) {
              /* If this is malloced */
              g_free (omx_buffer->pBuffer);
              omx_buffer->pBuffer = NULL;
            }
          }
        }
        OMX_FreeBuffer (port->core->omx_handle, port->port_index, omx_buffer);
      }
    }

    if (free) {
      g_free (port->buffers);
      port->num_buffers = 0;
      port->buffers = NULL;
    }
  }

  port->free_buffers = FALSE;
}

static void
omx_report_error (GOmxCore * core, OMX_U32 data_1)
{

  switch (data_1) {
    case OMX_ErrorBadParameter:
    case OMX_ErrorMbErrorsInFrame:
    case OMX_ErrorSeperateTablesUsed:
    case OMX_ErrorStreamCorrupt:
      GST_ELEMENT_ERROR (core->object,
          STREAM, DECODE, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorNotReady:
      GST_ELEMENT_ERROR (core->object,
          STREAM, FAILED, (NULL), (omx_error_to_str (data_1)));
      break;
    case OMX_ErrorFormatNotDetected:
    case OMX_ErrorUnsupportedSetting:
      GST_ELEMENT_ERROR (core->object,
          STREAM, FORMAT, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorNotImplemented:
      GST_ELEMENT_ERROR (core->object,
          STREAM, NOT_IMPLEMENTED, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorTimeout:
    case OMX_ErrorPortUnresponsiveDuringAllocation:
    case OMX_ErrorPortUnresponsiveDuringDeallocation:
    case OMX_ErrorPortUnresponsiveDuringStop:
      GST_ELEMENT_ERROR (core->object,
          RESOURCE, BUSY, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorInsufficientResources:
    case OMX_ErrorResourcesPreempted:
    case OMX_ErrorComponentSuspended:
    case OMX_ErrorPortUnpopulated:
    case OMX_ErrorDynamicResourcesUnavailable:
      GST_ELEMENT_ERROR (core->object,
          RESOURCE, FAILED, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorContentPipeCreationFailed:
    case OMX_ErrorResourcesLost:
      GST_ELEMENT_ERROR (core->object,
          RESOURCE, NOT_FOUND, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorContentPipeOpenFailed:
      GST_ELEMENT_ERROR (core->object,
          RESOURCE, OPEN_READ_WRITE, (NULL), (omx_error_to_str (data_1)));
      break;
    case OMX_ErrorUnderflow:
      GST_ELEMENT_ERROR (core->object,
          RESOURCE, READ, (NULL), (omx_error_to_str (data_1)));
      break;
    case OMX_ErrorOverflow:
      GST_ELEMENT_ERROR (core->object,
          RESOURCE, WRITE, (NULL), (omx_error_to_str (data_1)));
      break;
    case OMX_ErrorInvalidState:
    case OMX_ErrorComponentNotFound:
    case OMX_ErrorTunnelingUnsupported:
    case OMX_ErrorSameState:
    case OMX_ErrorUndefined:
      GST_ELEMENT_ERROR (core->object,
          LIBRARY, FAILED, (NULL), (omx_error_to_str (data_1)));
      break;
    case OMX_ErrorInvalidComponentName:
      GST_ELEMENT_ERROR (core->object,
          LIBRARY, INIT, (NULL), (omx_error_to_str (data_1)));
      break;
    case OMX_ErrorPortsNotCompatible:
    case OMX_ErrorVersionMismatch:
    case OMX_ErrorIncorrectStateTransition:
    case OMX_ErrorIncorrectStateOperation:
    case OMX_ErrorUnsupportedIndex:
    case OMX_ErrorBadPortIndex:
      GST_ELEMENT_ERROR (core->object,
          LIBRARY, SETTINGS, (NULL), (omx_error_to_str (data_1)));
      break;

    case OMX_ErrorNoMore:
      GST_ELEMENT_ERROR (core->object,
          LIBRARY, TOO_LAZY, (NULL), (omx_error_to_str (data_1)));
      break;

    default:
      GST_ELEMENT_ERROR (core->object,
          STREAM, FAILED, (NULL), (omx_error_to_str (data_1)));
  }
  return;
}


/*
 * OpenMAX IL callbacks.
 */

static OMX_ERRORTYPE
EventHandler (OMX_HANDLETYPE omx_handle,
    OMX_PTR app_data,
    OMX_EVENTTYPE event, OMX_U32 data_1, OMX_U32 data_2, OMX_PTR event_data)
{
  GOmxCore *core;

  core = (GOmxCore *) app_data;

  if (core->ehandler) {
    core->ehandler (GST_ELEMENT (core->object), event, data_1, data_2,
        event_data);
  }

  switch (event) {
    case OMX_EventCmdComplete:
    {
      OMX_COMMANDTYPE cmd;

      cmd = (OMX_COMMANDTYPE) data_1;

      GST_DEBUG_OBJECT (core->object, "OMX_EventCmdComplete: %d", cmd);

      switch (cmd) {
        case OMX_CommandStateSet:
          complete_change_state (core, data_2);
          break;
        case OMX_CommandFlush:
          g_sem_up (core->flush_sem);
          break;
        case OMX_CommandPortDisable:
          ((GOmxPort *) g_ptr_array_index (core->ports, data_2))->enabled =
              FALSE;
          g_sem_up (core->port_sem);
          break;
        case OMX_CommandPortEnable:
          ((GOmxPort *) g_ptr_array_index (core->ports, data_2))->enabled =
              TRUE;
          g_sem_up (core->port_sem);
          break;
        default:
          break;
      }
      break;
    }
    case OMX_EventBufferFlag:
    {
      GST_DEBUG_OBJECT (core->object, "OMX_EventBufferFlag");
      if (data_2 & OMX_BUFFERFLAG_EOS) {
        g_omx_core_set_done (core);
      }
      break;
    }
    case OMX_EventPortSettingsChanged:
    {
      GST_DEBUG_OBJECT (core->object, "OMX_EventPortSettingsChanged");
      /** @todo only on the relevant port. */
      core->event_data = event_data;
      core->event_type = data_2;
      if (core->settings_changed_cb) {
        GOmxPort *port;
        OMX_PARAM_PORTDEFINITIONTYPE param;
        port = ((GOmxPort *) g_ptr_array_index (core->ports, 1));

        if (data_2 == OMX_IndexParamPortDefinition && !port->eglbuffers) {
          port->reconfigure_buffers = TRUE;
          g_omx_port_pause (port);

          /*Wait till all valid buffers are rendered */
          g_sem_down (port->core->port_sem);
          GST_DEBUG_OBJECT (core->object,
              "OMX_EventPortSettingsChanged - All valid buffers are rendered.");

          /*Disable the decoder's output port */
          OMX_SendCommand (core->omx_handle, OMX_CommandPortDisable,
              port->port_index, NULL);
          port_free_buffers (port);
          g_sem_down (core->port_sem);
          port->reconfigure_buffers = FALSE;
          GST_DEBUG_OBJECT (core->object,
              "OMX_EventPortSettingsChanged - Decoder's output port is disabled.");

          /*Do decoder's output port setup */
          G_OMX_INIT_PARAM (param);
          param.nPortIndex = port->port_index;
          OMX_GetParameter (port->core->omx_handle,
              OMX_IndexParamPortDefinition, &param);
          port->num_buffers = param.nBufferCountActual;
          port->buffer_size = param.nBufferSize;
          port->buffers = g_new0 (OMX_BUFFERHEADERTYPE *, port->num_buffers);
          port->buffers_sent = 0;
          GST_DEBUG_OBJECT (core->object,
              "OMX_EventPortSettingsChanged - Decoder's output port has been setup.");

          core->settings_changed_cb (core);

          /*Enable the decoder's output port and allocate buffers at it */
          OMX_SendCommand (core->omx_handle, OMX_CommandPortEnable,
              port->port_index, NULL);
          port_allocate_buffers (port);
          g_sem_down (core->port_sem);
          port_start_buffers (port);
          GST_DEBUG_OBJECT (core->object,
              "OMX_EventPortSettingsChanged - Decoder's output port is enabled.");
        } else
          core->settings_changed_cb (core);
      }

      break;
    }
    case OMX_EventError:
    {
      core->omx_error = data_1;
      GST_ERROR_OBJECT (core->object, "unrecoverable error: %s (0x%lx)",
          omx_error_to_str (data_1), data_1);

      omx_report_error (core, data_1);

      /* component might leave us waiting for buffers, unblock */
      GST_ELEMENT_ERROR (core->object, STREAM, DECODE,
          (NULL), (omx_error_to_str (data_1)));

      g_omx_core_flush_start (core);
      /* unlock wait_for_state */
      g_mutex_lock (core->omx_state_mutex);
      g_cond_signal (core->omx_state_condition);
      g_mutex_unlock (core->omx_state_mutex);
      break;
    }
    default:
      break;
  }

  return OMX_ErrorNone;
}

static OMX_ERRORTYPE
EmptyBufferDone (OMX_HANDLETYPE omx_handle,
    OMX_PTR app_data, OMX_BUFFERHEADERTYPE * omx_buffer)
{
  GOmxCore *core;
  GOmxPort *port;

  core = (GOmxCore *) app_data;
  port = g_omx_core_get_port (core, omx_buffer->nInputPortIndex);

  GST_CAT_LOG_OBJECT (gstomx_util_debug, core->object, "omx_buffer=%p",
      omx_buffer);

  if (omx_buffer->pAppPrivate && (port->buffer_type == BUFFER_TYPE_NVBUFFER ||
          port->buffer_type == BUFFER_TYPE_NVRMBUFFER)) {
    GstBuffer *buf = (GstBuffer *) omx_buffer->pAppPrivate;

    omx_buffer->pAppPrivate = NULL;
    gst_buffer_unref (buf);
  }
  got_buffer (core, port, omx_buffer);

  return OMX_ErrorNone;
}

static OMX_ERRORTYPE
FillBufferDone (OMX_HANDLETYPE omx_handle,
    OMX_PTR app_data, OMX_BUFFERHEADERTYPE * omx_buffer)
{
  GOmxCore *core;
  GOmxPort *port;

  core = (GOmxCore *) app_data;
  port = g_omx_core_get_port (core, omx_buffer->nOutputPortIndex);
#ifdef GST_KPI_MEASURE
  if (G_UNLIKELY (port->first_buffer)) {
    if (omx_buffer->nOutputPortIndex == port->port_index) {
      if (port->start_timer != NULL) {
        g_timer_stop (port->start_timer);
        port->first_buffer = FALSE;
        g_print ("****time taken for receiving first buffer: %g secs\n",
            g_timer_elapsed (port->start_timer, NULL));
        g_timer_destroy (port->start_timer);
        port->start_timer = NULL;
      } else {
        g_print ("\nWARNING: port->start_timer is NULL\n");
        port->first_buffer = FALSE;
      }
    }
  }
#endif

  GST_CAT_LOG_OBJECT (gstomx_util_debug, core->object, "omx_buffer=%p",
      omx_buffer);
  got_buffer (core, port, omx_buffer);

  return OMX_ErrorNone;
}

static inline const char *
omx_state_to_str (OMX_STATETYPE omx_state)
{
  switch (omx_state) {
    case OMX_StateInvalid:
      return "invalid";
    case OMX_StateLoaded:
      return "loaded";
    case OMX_StateIdle:
      return "idle";
    case OMX_StateExecuting:
      return "executing";
    case OMX_StatePause:
      return "pause";
    case OMX_StateWaitForResources:
      return "wait for resources";
    default:
      return "unknown";
  }
}

static inline const char *
omx_error_to_str (OMX_ERRORTYPE omx_error)
{
  switch (omx_error) {
    case OMX_ErrorNone:
      return "None";

    case OMX_ErrorInsufficientResources:
      return
          "There were insufficient resources to perform the requested operation";

    case OMX_ErrorUndefined:
      return "The cause of the error could not be determined";

    case OMX_ErrorInvalidComponentName:
      return "The component name string was not valid";

    case OMX_ErrorComponentNotFound:
      return "No component with the specified name string was found";

    case OMX_ErrorInvalidComponent:
      return "The component specified did not have an entry point";

    case OMX_ErrorBadParameter:
      return "One or more parameters were not valid";

    case OMX_ErrorNotImplemented:
      return "The requested function is not implemented";

    case OMX_ErrorUnderflow:
      return "The buffer was emptied before the next buffer was ready";

    case OMX_ErrorOverflow:
      return "The buffer was not available when it was needed";

    case OMX_ErrorHardware:
      return "The hardware failed to respond as expected";

    case OMX_ErrorInvalidState:
      return "The component is in invalid state";

    case OMX_ErrorStreamCorrupt:
      return "Stream is found to be corrupt";

    case OMX_ErrorPortsNotCompatible:
      return "Ports being connected are not compatible";

    case OMX_ErrorResourcesLost:
      return "Resources allocated to an idle component have been lost";

    case OMX_ErrorNoMore:
      return "No more indices can be enumerated";

    case OMX_ErrorVersionMismatch:
      return "The component detected a version mismatch";

    case OMX_ErrorNotReady:
      return "The component is not ready to return data at this time";

    case OMX_ErrorTimeout:
      return "There was a timeout that occurred";

    case OMX_ErrorSameState:
      return
          "This error occurs when trying to transition into the state you are already in";

    case OMX_ErrorResourcesPreempted:
      return
          "Resources allocated to an executing or paused component have been preempted";

    case OMX_ErrorPortUnresponsiveDuringAllocation:
      return
          "Waited an unusually long time for the supplier to allocate buffers";

    case OMX_ErrorPortUnresponsiveDuringDeallocation:
      return
          "Waited an unusually long time for the supplier to de-allocate buffers";

    case OMX_ErrorPortUnresponsiveDuringStop:
      return
          "Waited an unusually long time for the non-supplier to return a buffer during stop";

    case OMX_ErrorIncorrectStateTransition:
      return "Attempting a state transition that is not allowed";

    case OMX_ErrorIncorrectStateOperation:
      return
          "Attempting a command that is not allowed during the present state";

    case OMX_ErrorUnsupportedSetting:
      return
          "The values encapsulated in the parameter or config structure are not supported";

    case OMX_ErrorUnsupportedIndex:
      return
          "The parameter or config indicated by the given index is not supported";

    case OMX_ErrorBadPortIndex:
      return "The port index supplied is incorrect";

    case OMX_ErrorPortUnpopulated:
      return
          "The port has lost one or more of its buffers and it thus unpopulated";

    case OMX_ErrorComponentSuspended:
      return "Component suspended due to temporary loss of resources";

    case OMX_ErrorDynamicResourcesUnavailable:
      return
          "Component suspended due to an inability to acquire dynamic resources";

    case OMX_ErrorMbErrorsInFrame:
      return "Frame generated macroblock error";

    case OMX_ErrorFormatNotDetected:
      return "Cannot parse or determine the format of an input stream";

    case OMX_ErrorContentPipeOpenFailed:
      return "The content open operation failed";

    case OMX_ErrorContentPipeCreationFailed:
      return "The content creation operation failed";

    case OMX_ErrorSeperateTablesUsed:
      return "Separate table information is being used";

    case OMX_ErrorTunnelingUnsupported:
      return "Tunneling is unsupported by the component";

    default:
      return "Unknown error";
  }
}
